/* QualityFnKernelDensityEstimator.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * @author Eric Eaton
 */

package ddpref;

import java.text.DecimalFormat;

import weka.estimators.KernelEstimator;

public class QualityFnKernelDensityEstimator extends KernelDensityEstimator {
	
	// set the max and min values to be extreme
	private double m_minValue = Double.POSITIVE_INFINITY;
	private double m_maxValue = Double.NEGATIVE_INFINITY;
	// store the precision
	private double m_precision = Double.NaN;

	public QualityFnKernelDensityEstimator(double precision) {
		super(precision);
		m_precision = precision;
	}

	@Override
	public void addValue(double data, double weight) {
		m_maxProbability = Double.NaN;  // set the maxProbability to be undetermined
		// update the max and min values
		m_minValue = (data < m_minValue) ? data : m_minValue;
		m_maxValue = (data > m_maxValue) ? data : m_maxValue;
		// add the value to the distribution
		super.addValue(data, weight);
	}

	@Override
	public double getProbability(double data) {
		double prob = super.getMaxProbability();
		if (prob == 0) {
			System.err.println("Problem in KernelDensityEstimator.generateDistribution(): m_maxProbability = 0.");
		}
		return super.getProbability(data) / prob;
	}
	
	
	public String toString() {
		
		double min = Math.rint(m_minValue / m_offset) * m_offset;
		double max = Math.rint(m_maxValue / m_offset) * m_offset;
		while (getProbability(min) > 0.01) { min -= m_precision; }
		while (getProbability(max) > 0.01) { max += m_precision; }
		
		if (true)
			return "KDE["+min+","+m_precision+","+max+"]";;
		
		String s = "";
		DecimalFormat df = new DecimalFormat("0.0000");
		//for (double i=0; i<100; i += 0.5) {
		//	s = s + df.format(getProbability(i)) + " ";
		//}
		//if (true) return s;
		
		//s = "KDE["+min+","+m_precision+","+max+"]:  ";
		
		int numProbabilities = (int) Math.ceil((max-min)/m_precision) + 1;
		double maxValue = min + ((numProbabilities-1) * m_precision);
		s = s = "KDE["+df.format(min)+","+df.format(m_precision)+","+df.format(maxValue)+"]:  ";
		df = new DecimalFormat("0.00");
		for (int i=0; i<numProbabilities; i++) {
			double value = min + (i*m_precision);
			s = s + df.format(getProbability(value))+" ";
		}
		//for (double i=min; i<=max; i += m_precision) {
		//	s = s + df.format(getProbability(i))+" ";
		//}
		
		//s += super.m_values;
		return s;
	}


}
