/* WekaUtils.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * @author Eric Eaton
 */

package ddpref;


import org.apache.commons.math.stat.descriptive.SummaryStatistics;
import org.apache.commons.math.stat.descriptive.SummaryStatisticsImpl;

import ddpref.experiments.ExperimentUtils;

import weka.core.Instance;
import weka.core.InstanceComparator;
import weka.core.Instances;

import ddpref.experiments.HungarianAlgorithm;


public class WekaUtils {

	public static Instances trimInstances(Instances instances, int numInstances) {
		Instances trimmedInstances = new Instances(instances);
		for (int i = trimmedInstances.numInstances() - 1; i >= numInstances; i--) {
			trimmedInstances.delete(i);
		}
		return trimmedInstances;
	}


	public static Instances mergeInstances(Instances instances1, Instances instances2) {
		if(instances1 == null || instances1.numInstances()==0)
			return instances2;
		if(instances2 == null || instances2.numInstances()==0)
			return instances1;
		if (!instances1.checkInstance(instances2.firstInstance()))
			throw new IllegalArgumentException("The instance sets are incompatible.");
		Instances mergedInstances = new Instances(instances1);
		Instances tempInstances = new Instances(instances2);
		for (int i = 0; i < tempInstances.numInstances(); i++) {
			mergedInstances.add(tempInstances.instance(i));
		}
		return mergedInstances;
	}
	
	/** Gets a subset of the given instances from [startIndex,endIndex], inclusive. */
	public static Instances getSubset(Instances instances, int startIndex, int endIndex) {
		if (startIndex < 0 || endIndex > (instances.numInstances()-1)) {
			throw new IllegalArgumentException("startIndex and endIndex out of range.");
		}
		Instances newInstances = new Instances(instances,0);
		for (int i=startIndex; i<=endIndex; i++) {
			newInstances.add(instances.instance(i));
		}
		return newInstances;
	}
	
	public static Instances getNonMembers(Instances data, Instances instancesToExclude) {
		Instances nonMembers = new Instances(data,0);
		for (int i=0; i<data.numInstances(); i++) {
			// get an instance
			Instance tempInstance = data.instance(i);
			// add it if it's not in the instances to exclude
			if (!isMember(tempInstance, instancesToExclude)) {
				nonMembers.add(tempInstance);
			}
		}
		return nonMembers;
	}
	
	private static InstanceComparator comparator = new InstanceComparator();
	
	/** Determines if the specified instance is a member of the instances.
	 * @param instance
	 * @param instances
	 * @return <code>true</code> if instance is in instances, <code>false</code> otherwise.
	 */
	public static boolean isMember(Instance instance, Instances instances) {
		for (int instIdx=0; instIdx<instances.numInstances(); instIdx++) {
			if (comparator.compare(instance, instances.instance(instIdx)) == 0) {
				return true;
			}
		}
		return false;
	}


	/**
	 * Converts an instance to a feature vector excluding the class attribute.
	 * @param instance The instance.
	 * @return A vector representation of the instance excluding the class attribute
	 */
	public static double[] instanceToFeatureVector(Instance instance) {
		double[] vector = new double[(instance.classIndex() != -1) ? instance.numAttributes() - 1 : instance.numAttributes()];
		double[] instanceDoubleArray = instance.toDoubleArray();
		int attIdx = 0;
		for (int i = 0; i < vector.length; i++) {
			if (i == instance.classIndex()) {
				attIdx++;
			}
			vector[i] = instanceDoubleArray[attIdx++];
		}
		return vector;
	}


	/**
	 * Converts a set of instances to an array of vectors
	 * @param instances The set of instances.
	 * @return The array of feature vectors.
	 */
	public static double[][] instancesToFeatureVectors(Instances instances) {
		double[][] vectors = new double[instances.numInstances()][];
		for (int instIdx = 0; instIdx < instances.numInstances(); instIdx++) {
			vectors[instIdx] = instanceToFeatureVector(instances.instance(instIdx));
		}
		return vectors;
	}


	public static SummaryStatistics getPredictionStats(Instances data, int[] predictedClasses) {
		SummaryStatistics stats = new SummaryStatisticsImpl();
		for (int instIdx = 0; instIdx < data.numInstances(); instIdx++) {
			Instance instance = data.instance(instIdx);
			if (instance.classValue() == (double) predictedClasses[instIdx])
				stats.addValue(1);
			else
				stats.addValue(0);
		}
		return stats;
	}
	
	/** Computes the precision of the sample data against the target data */
	public static double precision(Instances sampleData, Instances targetData) {
		double relevantAndRetrieved = 0;
		// loop through the sample data
		for (int instIdx = 0; instIdx < sampleData.numInstances(); instIdx++) {
			Instance instance = sampleData.instance(instIdx);
			if (WekaUtils.isMember(instance, targetData)) {
				relevantAndRetrieved++;
			}
		}
		// compute the precision
		double precision = relevantAndRetrieved / sampleData.numInstances();
		return precision;
	}
	
	
	/** Computes the recall of the sample data against the target data */
	public static double recall(Instances sampleData, Instances targetData) {
		double relevantAndRetrieved = 0;
		// loop through the sample data
		for (int instIdx = 0; instIdx < sampleData.numInstances(); instIdx++) {
			Instance instance = sampleData.instance(instIdx);
			if (WekaUtils.isMember(instance, targetData)) {
				relevantAndRetrieved++;
			}
		}
		// compute the recall
		double recall = relevantAndRetrieved / targetData.numInstances();
		return recall;
	}
	
	/** Computes the fmeasure of the sample data against the target data */
	public static double fmeasure(Instances sampleData, Instances targetData) {
		double relevantAndRetrieved = 0;
		// loop through the sample data
		for (int instIdx = 0; instIdx < sampleData.numInstances(); instIdx++) {
			Instance instance = sampleData.instance(instIdx);
			if (WekaUtils.isMember(instance, targetData)) {
				relevantAndRetrieved++;
			}
		}
		// shortcut for the sizes of the two sets being the same
		double fmeasure;
		if (sampleData.numInstances() == targetData.numInstances()) {
			// the precision and recall are the same for the same size sets, so their harmonic mean is just the precision (or recall)
			fmeasure = relevantAndRetrieved / sampleData.numInstances();
		} else {
			// compute the precision
			double precision = relevantAndRetrieved / sampleData.numInstances();
			// compute the recall
			double recall = relevantAndRetrieved / targetData.numInstances();
			// compute the f-measure
			if ((precision + recall) == 0) {
				fmeasure = 0;
			} else {
				fmeasure = 2 * precision * recall / (precision + recall);
			}
		}
		if (Double.isNaN(fmeasure)) {
			System.out.println("NaN!!!!");
		}
		return fmeasure;
	}
	
	/** Gets the Euclidean distance between two data sets. */
	public static double euclideanDistance(Instances data1, Instances data2) {
		// make certain the sizes of the two sets are the same
		if (data1.numInstances() != data2.numInstances()) {
			System.err.println("Error: Can't compute Euclidean distance; data1.size = "+data1.numInstances()+" != "+data2.numInstances()+" = data2.size");
			return Double.NaN;
		}
		// identify different items in the set
		Instances d1R = getNonMembers(data1,data2);
		Instances d2R = getNonMembers(data2,data1);
		int sizeD1R = d1R.numInstances();
		int sizeD2R = d2R.numInstances();
		// if their sizes are different, there's a serious problem
		if (sizeD1R != sizeD2R)
			throw new IllegalStateException("The sizes of D1R and D2R differ.  Problem in WekaUtils.getNonMembers()?");
		
		// if both sets are empty, then the euclidean distance is 0
		if (sizeD1R == 0) {
			return 0;
		}
		
		// otherwise, compute the Euclidean distance
		// create a matrix of the distance between every pair of items
		double[][] dists = new double[sizeD1R][sizeD2R];
		for (int i=0; i<sizeD1R; i++) {
			Instance inst1 = d1R.instance(i);
			for (int j=0; j<sizeD2R; j++) { 
				Instance inst2 = d2R.instance(j);
				dists[i][j] = euclideanDistance(inst1,inst2);
			}
		}
		// find the optimal pairing of the instances to minimize the distance
		// this is just a simple task assignment problem, which we can use the hungarian algorithm for
		int[][] assignments = HungarianAlgorithm.hgAlgorithm(dists,"min");
		double euclideanDistance = 0;
		for (int[] assignment : assignments)
			euclideanDistance += dists[assignment[0]][assignment[1]];
		// return this minimum summed distance as the euclidean distance
		return euclideanDistance;
		
	}
	
	public static double euclideanDistance(Instance inst1, Instance inst2) {
		double dist = 0;
		if (inst1.numAttributes() != inst2.numAttributes()) {
			throw new IllegalArgumentException("inst1 and inst2 must have the same number of attributes.");
		}
		for (int attIdx=0; attIdx<inst1.numAttributes(); attIdx++) {
			double value = inst1.value(attIdx) - inst2.value(attIdx);
			dist += Math.pow(value,2);
		}
		return Math.sqrt(dist);
	}
	


	public static double[] vectorAdd(double[] v1, double[] v2) {
		double[] result = new double[v1.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = v1[i] + v2[i];
		}
		return result;
	}


	public static double[] vectorSubtract(double[] v1, double[] v2) {
		double[] result = new double[v1.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = v1[i] - v2[i];
		}
		return result;
	}


	public static double[] vectorMultiply(double s, double[] v) {
		double[] result = new double[v.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = s * v[i];
		}
		return result;
	}


	public static double vectorL2Norm(double[] v) {
		double norm = 0;
		for (double d : v) {
			norm += Math.pow(d, 2);
		}
		return Math.sqrt(norm);
	}


	public static double[] vectorNormalize(double[] v) {
		double[] result = new double[v.length];
		double sum = 0;
		for (double d : v) {
			sum += d;
		}
		for (int i = 0; i < v.length; i++) {
			result[i] = v[i] / sum;
		}
		return result;
	}
	
}
